# -*- coding: utf-8 -*-
"""
Created on Tue Sep 13 13:44:52 2016

@author: albert.kurucz
"""
from __future__ import absolute_import
# STANDARD IMPORTS
import threading
import sys
import logging
import argparse
from tempfile import NamedTemporaryFile
from time import time
from time import sleep
from os import remove
from os.path import isfile
import bz2
from binascii import hexlify
# NON STANDARD IMPORTS
from intelhex import IntelHex
from intelhex import NotEnoughDataError
from serial import Serial
import serial.tools.list_ports as lp
from Crypto.Cipher import ARC4
try:
    from .snapavr_version import __version__
    assert __version__
except (ImportError, ValueError):
    __version__ = "UNKNOWN"
try:
    from .snapavr_version import __revision__
    assert __revision__
except (ImportError, ValueError):
    __revision__ = "UNKNOWN"

DEFAULT_TIMEOUT = 30.0   # 30 seconds


class Timeout(Exception):
    """Timeout while waiting for bootloader"""
    pass


class AddressResponseNotReceived(Exception):
    """'A' (set address) command returned unexpected response"""
    pass


class ExitResponseNotReceived(Exception):
    """'E' (exit bootloader) command returned unexpected response"""
    pass


class InvalidBlockSizeResponse(Exception):
    """'b' (block size read) command returned unexpected response"""
    pass


class SerialPortIsNotFound(Exception):
    """Serial port is not found"""
    pass


class TooManySerialPortsFound(Exception):
    """Too many serial ports found"""
    pass


class ChecksumError(Exception):
    """'B' (block write) command returned unexpected checksum"""
    pass


class SignatureMismatch(Exception):
    """'s' (signature read) command returned unexpected signature"""
    pass


class BlockSizeReadFailed(Exception):
    """'b' (block size read) command failed"""
    pass


class SignatureReadFailed(Exception):
    """'s' (signature read) command failed"""
    pass


class IdReadFailed(Exception):
    """'I' (id read) command failed"""
    pass


def compute_checksum(data):
    """compute checksum of data"""
    computed = 0
    for char in data:
        computed += ord(char)
    computed &= 0xffff
    return computed


class AvrBootloader(Serial):
    """The bootloader class based of Serial"""
    # pylint: disable=R0901
    def __init__(self, serial_port, waiting=True, rts_reset=True):
        """initialize the bootloader class"""
        logging.info('opening serial port: %s', serial_port)
        Serial.__init__(
            self,
            serial_port,
            115200
        )
        self.waiting = waiting
        self.rts_reset = rts_reset
        if self.rts_reset is not None:
            logging.info('reset')
            self.setRTS(self.rts_reset)
        self.timeout = 0.1
        self.thread = threading.Thread(target=self.start)
        self.thread.start()

    def start(self):
        """start waiting for bootloader protocol"""
        logging.info('starting...')
        if self.rts_reset is not None:
            logging.info('unreset')
            self.setRTS(not self.rts_reset)
        while self.waiting:
            chin = self.read(1)
            if chin == '\xf9':
                self.write('\xf6')
                self.flush()
                self.waiting = False

    def wait(self, timeout=DEFAULT_TIMEOUT):
        """wait for bootloader protocol to start"""
        end_time = time() + timeout
        while self.waiting:
            if time() > end_time:
                self.waiting = False
                self.thread.join()
                self.close()
                raise Timeout()
            sleep(0.1)

    def read_id(self):
        """read booloader software identification"""
        self.flushInput()
        self.write('I')
        self.flush()
        id_read = self.read(3)
        if len(id_read) == 3:
            logging.info('version: %d', ord(id_read[0]))
            return id_read
        raise SignatureReadFailed()

    def read_signature(self):
        """read chip signation"""
        self.flushInput()
        self.write('s')
        self.flush()
        signature_read = self.read(3)
        if len(signature_read) == 3:
            logging.info('sig1: %s', hex(ord(signature_read[2])))
            logging.info('sig2: %s', hex(ord(signature_read[1])))
            logging.info('sig3: %s', hex(ord(signature_read[0])))
            return signature_read
        raise IdReadFailed()

    def set_address(self, address):
        """set the address for the nex block read/write operation"""
        # logging.info('address: 0x%x', address)
        self.flushInput()
        self.write('A')
        self.write(chr(0xff & (address >> 9)))
        self.write(chr(0xff & (address >> 1)))
        self.flush()
        rec = self.read(1)
        if rec != '\r':
            raise AddressResponseNotReceived(
                'address: 0x%x, received: %s' % (address, repr(rec))
            )

    def get_block_size(self):
        """get the block size"""
        self.flushInput()
        self.write('b')
        self.flush()
        blocksize_read = self.read(3)
        if len(blocksize_read) == 3:
            if blocksize_read[0] != 'Y':
                raise InvalidBlockSizeResponse()
            blocksize = (
                ord(blocksize_read[1]) << 8
            ) | ord(blocksize_read[2])
            logging.info('blocksize: %d', blocksize)
            return blocksize
        raise BlockSizeReadFailed()

    def write_hex(self, intel_hex, memory_type='F'):
        """write multiple blocks of data, defined in a hex file"""
        # fist prepare the hex segments
        for segment_start, segment_end in intel_hex.segments():
            while segment_start % 256:
                segment_start -= 1
                try:
                    intel_hex.gets(segment_start, 1)
                except NotEnoughDataError:
                    intel_hex[segment_start] = 255
        # write the hex segments after prepared
        for segment_start, segment_end in intel_hex.segments():
            self.write_segment(
                intel_hex.gets(
                    segment_start,
                    segment_end - segment_start
                ),
                segment_start,
                memory_type,
            )

    def write_segment(self, data, segment_start, memory_type='F'):
        """write a block of data, size not limited"""
        offset = 0
        segment_len = len(data)
        self.set_address(segment_start)
        while segment_len > 0:
            write_len = 256 if segment_len > 256 else segment_len
            try:
                self.write_block(
                    data[offset:offset + write_len],
                    memory_type
                )
            except ChecksumError:
                logging.error('checksum error at 0x%x', segment_start)
                raise
            offset += write_len
            segment_start += write_len
            segment_len -= write_len

    def write_block(self, data, memory_type='F'):
        """write a block of data, size not larger than 256"""
        data_len = len(data)
        self.flushInput()
        self.write('B')
        self.write(chr(0xff & (data_len >> 8)))
        self.write(chr(0xff & data_len))
        self.write(memory_type)
        self.write(data)
        self.flush()
        checksum_read = self.read(2)
        computed = compute_checksum(data)
        if len(checksum_read) == 2:
            checksum = (
                ord(checksum_read[0]) << 8
            ) | ord(checksum_read[1])
            if checksum != computed:
                logging.error(
                    'checksum: 0x%x (0x%x)',
                    checksum,
                    computed
                )
                raise ChecksumError(
                    'checksum: 0x%x, computed: 0x%x' % (
                        checksum,
                        computed
                    )
                )

    def read_block(self, data_len, memory_type='F'):
        """read a block of data"""
        self.flushInput()
        self.write('g')
        self.write(chr(0xff & (data_len >> 8)))
        self.write(chr(0xff & data_len))
        self.write(memory_type)
        self.flush()
        # data_len = 2 * data_len
        data_read = self.read(data_len)
        actual_data_len = len(data_read)
        if actual_data_len != data_len:
            logging.error(
                'read length requested: %d, actual: %d',
                data_len,
                actual_data_len
            )
        return data_read

    def exit_bootloader(self):
        """exit bootloader"""
        self.flushInput()
        self.write('E')
        self.flush()
        response_read = self.read(1)
        if response_read != '\r':
            raise ExitResponseNotReceived()


def snap_avr(
        serial_port=None,
        usb_vid=None,
        usb_pid=None,
        usb_serial_number=None,
        usb_location=None,
        timeout=DEFAULT_TIMEOUT,
        do_not_exit_bootloader=False,
        sfi_file=None,
        hex_file=None,
        spy_file=None,
        magic=None,
        signature='\x04\x97\x1e',
):
    """Entry point for command line, whith arguments parsed already"""
    # pylint: disable=R0912,R0913,R0914,R0915
    start_time = time()
    bl_time = None
    logging.info("timeout: %.3f", timeout)
    logging.info("do_not_exit_bootloader: %s", do_not_exit_bootloader)
    if serial_port:
        logging.info("serial_port: %s", serial_port)
    else:
        if usb_vid:
            logging.info("usb_vid: %d", usb_vid)
        if usb_pid:
            logging.info("usb_pid: %d", usb_pid)
        if usb_serial_number:
            logging.info("usb_serial_number: %s", usb_serial_number)
        if usb_location:
            logging.info("usb_location: %s", usb_location)
        com_ports = lp.comports()
        my_com_ports = []
        for com_port in reversed(com_ports):
            logging.info(
                "%s %d %d %s %s",
                com_port.device,
                com_port.vid,
                com_port.pid,
                repr(com_port.serial_number),
                repr(com_port.location),
            )
            if usb_vid:
                if usb_vid != com_port.vid:
                    continue
            if usb_pid:
                if usb_pid != com_port.pid:
                    continue
            if usb_serial_number:
                if usb_serial_number != com_port.serial_number:
                    continue
            if usb_location:
                if usb_location != com_port.location:
                    continue
            my_com_ports.append(com_port)
        if not my_com_ports:  # if list is empty!
            raise SerialPortIsNotFound()
        if len(my_com_ports) > 1:
            raise TooManySerialPortsFound()
        serial_port = my_com_ports[0].device
    bootloader = None
    try:
        bootloader = AvrBootloader(serial_port)
        bootloader.wait(timeout)
        signature_read = bootloader.read_signature()
        if signature_read != signature:
            logging.error(
                'signature read: %s, expected: %s',
                hexlify(signature_read),
                hexlify(signature),
            )
            raise SignatureMismatch(
                'signature read: %s, expected: %s' %
                (hexlify(signature_read), hexlify(signature))
            )
        bl_time = time()
        bootloader.read_id()
        if sfi_file is not None:
            with open(sfi_file, 'rb') as sfi_stream:
                sfi_text = sfi_stream.read()
                # sfi files are compressed
                hex_text = bz2.decompress(sfi_text)
            try:
                with NamedTemporaryFile(delete=False) as temp:
                    temp.write(hex_text)
                    temp_file = temp.name
                bootloader.write_hex(IntelHex(temp_file))
            finally:
                if isfile(temp_file):
                    try:
                        remove(temp_file)
                    except OSError:
                        logging.error(
                            'cannot remove temporary file: %s',
                            temp_file
                        )
        if hex_file is not None:
            bootloader.write_hex(IntelHex(hex_file))
        if spy_file is not None:
            with open(spy_file, 'rb') as spy_stream:
                spy_content = spy_stream.read()
            algo = ARC4.new("SynapseExport3.14159265359")
            bootloader.write_segment(
                bz2.decompress(algo.decrypt(spy_content)),
                0xF00,
            )
        if magic is not None:
            bootloader.write_segment(magic, 0x2F0)
        if (hex_file is None) and (spy_file is None) and (magic is None):
            # by default perform a magical full erase
            bootloader.write_segment("RMDEMGKBF", 0x2F0)
        if do_not_exit_bootloader is False:
            bootloader.exit_bootloader()
    finally:
        if bootloader is not None:
            bootloader.close()
    finish_time = time()
    logging.info(
        'total runtime: %.3f',
        finish_time - start_time,
    )
    if bl_time is not None:
        logging.info(
            'programming time: %.3f',
            finish_time - bl_time,
        )


def snap_avr_main():
    """Entry point for command line"""
    logging.basicConfig(level=logging.INFO)
    logging.info('snapavr version: %s', __version__)
    logging.info('snapavr revision: %s', __revision__)
    logging.info(str(sys.version_info))
    parser = argparse.ArgumentParser()
    parser.add_argument("--serial_port", help="serial port")
    parser.add_argument("--usb_location", help="usb location")
    parser.add_argument("--usb_pid", type=int, help="usb product id")
    parser.add_argument("--usb_vid", type=int, help="usb vendor id")
    parser.add_argument("--usb_serial_number", help="usb serial number")
    parser.add_argument("--sfi_file", help="sfi file")
    parser.add_argument("--hex_file", help="hex file")
    parser.add_argument("--spy_file", help="spy file")
    parser.add_argument("--magic", help="magic")
    parser.add_argument(
        "--do_not_exit_bootloader",
        dest='do_not_exit_bootloader',
        action='store_true',
        help="do not exit bootloader",
    )
    parser.add_argument(
        "--timeout",
        type=float,
        default=DEFAULT_TIMEOUT,
        help="timeout to engage bootloder",
    )
    args = parser.parse_args()
    snap_avr(
        serial_port=args.serial_port,
        usb_vid=args.usb_vid,
        usb_pid=args.usb_pid,
        usb_serial_number=args.usb_serial_number,
        usb_location=args.usb_location,
        timeout=args.timeout,
        do_not_exit_bootloader=args.do_not_exit_bootloader,
        sfi_file=args.sfi_file,
        hex_file=args.hex_file,
        spy_file=args.spy_file,
        magic=args.magic,
    )
    logging.info('done')
